# This code is copyrighted by Stephan Richter and
# Thomas Foerster in September, 2001.
#
# written by Stephan Richter
# Last updated: October 4, 2001
import string


class Operator:

    reprs = []

    def __init__(self, lhs, rhs):
        ''' '''
        if type(lhs) is type(()):
            self.lhs = lhs[1]
        else:
            self.lhs = lhs

        if type(rhs) == type(()):
            self.rhs = rhs[1]
        else:
            self.rhs = rhs


    def __getitem__(self, index):
        ''' '''
        if index == 0:
            return self.lhs
        elif index == 1:
            return self.rhs
        else:
            return None
        

    def __str__(self):
        ''' '''
        return str(self.lhs) + ' ' + self.reprs[0] + ' ' + str(self.rhs)


    def _get_values(self, index):
        ''' '''
        result = []
        for side_str in ('lhs', 'rhs'):
            side = getattr(self, side_str)

            if hasattr(side, 'is_identifier'):
                side = side.execute(index)

            # YIKES!!! FIX, FIX, FIX ...
            if callable(side):
                try:
                    side = side(index)
                except:
                    try:
                        side = side()
                    except:
                        pass

            result.append(side)

        return tuple(result)


    def execute(self, index):
        ''' '''
        pass


    def __call__(self, index):
        return self.execute(index)



# ** Start of Boolean Operators Implementation **

class BooleanOperator(Operator):
    pass


class Equal(BooleanOperator):

    reprs = ['==', 'equal']

    def execute(self, index):
        lhs = get_identifier_name(self.lhs)
        rhs = get_identifier_name(self.rhs)
        return apply(index, (), {lhs: rhs})


class NotEqual(BooleanOperator):

    reprs = ['<>', '!=', 'nequal']

    def execute(self, index):
        lhs = get_identifier_name(self.lhs)
        rhs = get_identifier_name(self.rhs)
        return lhs != rhs


class GreaterThanEqual(BooleanOperator):

    reprs = ['>=', 'gtequal']

    def execute(self, index):
        lhs = get_identifier_name(self.lhs)
        rhs = get_identifier_name(self.rhs)
        return apply(index, (), {lhs: rhs, lhs+'_usage': 'range:min'})


class LessThanEqual(BooleanOperator):

    reprs = ['<=', 'ltequal']

    def execute(self, index):
        lhs = get_identifier_name(self.lhs)
        rhs = get_identifier_name(self.rhs)
        return apply(index, (), {lhs: rhs, lhs+'_usage': 'range:max'})


class GreaterThan(BooleanOperator):

    reprs = ['>', 'greater']

    def execute(self, index):
        lhs = get_identifier_name(self.lhs)
        rhs = get_identifier_name(self.rhs)
        return apply(index, (), {lhs: rhs, lhs+'_usage': 'range:min'})


class LessThanEqual(BooleanOperator):

    reprs = ['<', 'lesser']

    def execute(self, index):
        lhs = get_identifier_name(self.lhs)
        rhs = get_identifier_name(self.rhs)
        return apply(index, (), {lhs: rhs, lhs+'_usage': 'range:max'})


class And(BooleanOperator):

    reprs = ['&', 'and']

    def execute(self, index):
        lhs, rhs = self._get_values(index)        
        rhs_uids = map(lambda x: x.data_record_id_, rhs) 
        result = []
        for obj in lhs:
            if obj.data_record_id_ in rhs_uids:
                result.append(obj)
        return result


class Or(BooleanOperator):

    reprs = ['|', 'or']

    def execute(self, index):
        lhs, rhs = self._get_values(index)        
        result = []
        for obj in lhs+rhs:
            if obj.data_record_id_ not in map(lambda x: x.data_record_id_, result):
                result.append(obj)
        return result


class Xor(BooleanOperator):

    reprs = ['^', 'xor']

    def execute(self, index):
        lhs, rhs = self._get_values(index)        
        lhs_uids = map(lambda x: x.data_record_id_, lhs) 
        rhs_uids = map(lambda x: x.data_record_id_, rhs) 
        result = []

        for obj in lhs:
            if obj.data_record_id_ not in rhs_uids:
                result.append(obj)

        for obj in rhs:
            if obj.data_record_id_ not in lhs_uids:
                result.append(obj)

        return result


class Not(BooleanOperator):

    reprs = ['!', 'not']

    def execute(self, index):
        rhs = self._get_values(index)[1]        
        rhs_uids = map(lambda x: x.data_record_id_, rhs) 
        all = index()
        result = []
        for obj in all:
            if obj.data_record_id_ not in rhs_uids:
                result.append(obj)
        return result



def get_identifier_name(elem):
    if hasattr(elem, 'attr'):
        return elem.attr
    else:
        return elem


def convert_operator(operator):

    if not operator or not hasattr(operator, 'is_operator'):
        return operator

    if operator.lhs:
        lhs = convert_operator(operator.lhs)

    if operator.rhs:
        rhs = convert_operator(operator.rhs)

    if operator.reprs[0] == '==':
        return Equal(lhs, rhs)
    elif operator.reprs[0] == '<>':
        return NotEqual(lhs, rhs)
    elif operator.reprs[0] == '>=':
        return GreaterThanEqual(lhs, rhs)
    elif operator.reprs[0] == '<=':
        return LessThanEqual(lhs, rhs)
    elif operator.reprs[0] == '>':
        return GreaterThan(lhs, rhs)
    elif operator.reprs[0] == '<':
        return LessThan(lhs, rhs)
    elif operator.reprs[0] == '&':
        return And(lhs, rhs)
    elif operator.reprs[0] == '|':
        return Or(lhs, rhs)
    elif operator.reprs[0] == '^':
        return Xor(lhs, rhs)
    elif operator.reprs[0] == '!':
        return Not(lhs, rhs)
    else:
        return operator
