# This code is copyrighted by Stephan Richter and
# Thomas Foerster in September, 2001.
#
# written by Stephan Richter
# Last updated: October 4, 2001
"""grammar generation stuff for oql.

This module does not bind any rule semantics, it
just generates the parser data structures.
"""

MARSHALFILE = INSTANCE_HOME + "/var/oql.mar"


### interpretation functions and regexen for terminals

from DateTime.DateTime import DateTime
from string import letters, digits
import string

def Any(strlist):
    regstr = '\('+string.join(strlist, '\)\|\(')+'\)'
    return regstr
        
# python style comments
commentre = "#.*" 

## python terminals
identre = "[%s_][%s%s_]*" % (letters, letters, digits)

# literals
sstrre = "'[^\n\\']*'"
dstrre = '"[^\n\\"]*"'

stringre = Any([sstrre, dstrre])
## XXX no long strings and no raw string specifiers yet (necessary?)


# DateTime handler
single_datetime_re = '\(%s\):dt' %(sstrre)
double_datetime_re = '\(%s\):dt' %(dstrre)

datetimere =  single_datetime_re #Any([single_datetime_re, double_datetime_re])

def makeDateTime(str):
    return DateTime(str[1:-4])

# number terminals
intre = "\([1-9][0-9]*\)\|\(0\)"
## XXX no hex and octal integers yet
lintre = "\(%s\)[lL]" % (intre)

zerore = "\(%s\)\." % (intre)
fracre = "\(%s\)?\.[0-9]+"
exponentre = "[eE][\+\-][0-9]*"
pointfloatre = Any([zerore,fracre])
expfloatre = '\(' + Any([intre, pointfloatre]) +'\)' + exponentre

floatre = Any([expfloatre, pointfloatre])

def evalfn(str):
    return eval(str)

def stripfn(str):
    return str[1:-1]

# We need to make the difference between an identifier and a string.
# This class will allow us to do that.
class Identifier:

    is_identifier = 1

    def __init__(self, str):
        self.attr = str


    def __str__(self):
        return self.attr


    def execute(self, object):
        return getattr(object, self.attr, None)


### Declaration of the terminal symbols
def DeclareTerminals(Grammar):
    """ Declaration of terminal symbols """

    ## NOTE: Ordering of the definitions is important, kwParsing seems to use first
    ## matching rule, not longest match
    
    Grammar.Addterm("identifier", identre, Identifier)
    Grammar.Addterm("datetime", datetimere, makeDateTime)
    Grammar.Addterm("string_literal", stringre, stripfn)
    Grammar.Addterm("floatnumber", floatre, evalfn)
    Grammar.Addterm("longinteger", lintre, evalfn)
    Grammar.Addterm("integer", intre, evalfn)
    Grammar.Addterm("plusequal", "+=", None)
    Grammar.Addterm("minusequal", "-=", None)
    Grammar.Addterm("mulequal", "*=", None)
    Grammar.Addterm("divequal", "/=", None)
    Grammar.Addterm("modequal", "%=", None)
    Grammar.Addterm("plus", "\+", None)
    Grammar.Addterm("minus", "\-", None)
    Grammar.Addterm("dast", "\*\*", None)
    Grammar.Addterm("ast", "\*", None)
    Grammar.Addterm("slash", "/", None)
    Grammar.Addterm("percent", "%", None)
    Grammar.Addterm("dless", "<<", None)
    Grammar.Addterm("dgreater", ">>", None)
    Grammar.Addterm("gtequal", ">=", None)
    Grammar.Addterm("ltequal", "<=", None)
    Grammar.Addterm("nequal", "<>", None)
    Grammar.Addterm("nequal", "!=", None)
    Grammar.Addterm("equal", "==", None)
    Grammar.Addterm("assign", "\=", None)
    Grammar.Addterm("less", "<", None)
    Grammar.Addterm("greater", ">", None)
    Grammar.Addterm("amper", "&", None)
    Grammar.Addterm("pipe", "|", None)
    Grammar.Addterm("circumflex", "\^", None)
    Grammar.Addterm("tilde", "~", None)
    
### Build Function of the grammar
def BuildOQL(filename=MARSHALFILE):
    """ Builds an OQL Grammar from descriptions and stores it in a binary file.
    The filename may be given as argument (DEFAULT " oql.mar"). """
    
    import kjParseBuild
    from oqlgram import oqlrules, nonterms, keywords, puncts
    from oqlsem import BindRules
    
    # Grammar description --> see oqlgram.py for details
    OQLG = kjParseBuild.NullCGrammar()
    OQLG.SetCaseSensitivity(0)
    DeclareTerminals(OQLG)
    OQLG.Keywords(keywords)
    OQLG.punct(puncts)
    OQLG.Nonterms(nonterms)
    OQLG.comments([commentre])
    OQLG.Declarerules(oqlrules)

    print "compiling Grammar..."
    OQLG.Compile()
    
    print "dumping to", filename
    outfile = open(filename, "wb")
    OQLG.MarshalDump(outfile)
    outfile.close()

    BindRules(OQLG)
    
    return OQLG
    
def ReloadOQL(filename=MARSHALFILE):
    """ Reload a grammar from filename (Default "oql.mar"). NOTE: This
    function does not bind any interpretation functions."""
    import kjParser
    from oqlsem import BindRules
    
    try:
        infile = open(filename, "rb")
        OQLG = kjParser.UnMarshalGram(infile)
        infile.close()
        DeclareTerminals(OQLG)
        BindRules(OQLG)
    except IOError:
        # file not found ->
        # automatically rebuild grammar
        OQLG = BuildOQL(filename);
        
    
    return OQLG
    

    
    

    
